# frozen_string_literal: true

require "is/inspectable"

require_relative "../../is/async"

module Core
  module Async
    class Collection
      class << self
        include Is::Async

        # [public] Builds a collection from an enumerable object.
        #
        def build(object)
          aware do
            errored, stopped = false

            values = object.map { |value|
              break if errored || stopped

              async do
                if block_given?
                  yield value
                else
                  value
                end
              rescue LocalJumpError
                stopped = true
              rescue => error
                errored = true
                raise error
              end
            }

            new(values) if values
          end
        end
      end

      include Enumerable
      include Is::Async
      include Is::Inspectable

      def initialize(values = [])
        unless values.respond_to?(:each)
          raise ArgumentError, "values are not enumerable"
        end

        @values = values
      end

      # [public] Adds a value to the collection.
      #
      def <<(value)
        @values << value
      end
      alias_method :add, :<<

      # [public] Yields each resolved value.
      #
      def each
        return to_enum(:each) unless block_given?

        @values.each do |value|
          yield resolve(value)
        end
      end
    end
  end
end
