# frozen_string_literal: true

require "is/inspectable"

module Core
  module Async
    # [public] Represents a future result.
    #
    class Future
      include Is::Inspectable
      inspects without: [:@task]

      def initialize(task)
        @task = task
        @error = nil
      end

      # [public] Wait on the future to resolve, returning self.
      #
      def wait
        unless @error
          resolve
        end

        self
      end

      # [public] Attempt to cancel the future, returns true if successful.
      #
      def cancel
        if pending?
          @task.stop
        end

        self
      end

      # [public] Return the result, blocking until available.
      #
      def result
        @error || @result ||= resolve
      end

      private def resolve
        wait_all(@task)

        resolve_value(@task.result)
      rescue UncaughtThrowError => error
        throw error.tag, error.value
      rescue => error
        @error = error

        raise error
      end

      # [public] Return any error that occurred without re-raising, blocking until available.
      #
      def error
        @error ||= get_error
      end

      private def get_error
        result
        nil
      rescue => error
        error
      end

      # [public] Return the status; one of :pending, :failed, or :complete.
      #
      def status
        if defined?(@status)
          @status
        else
          status = find_status
          if terminal_status?(status)
            @status = status
          end

          status
        end
      end

      private def find_status
        case @task.status
        when :running
          :pending
        when :stopped
          :canceled
        when :failed
          :failed
        when :complete
          :complete
        end
      end

      private def terminal_status?(status)
        status == :failed || status == :complete
      end

      # [public] Return `true` if pending.
      #
      def pending?
        status == :pending
      end

      # [public] Return `true` if failed.
      #
      def failed?
        status == :failed
      end

      # [public] Return `true` if canceled.
      #
      def canceled?
        status == :canceled
      end

      # [public] Return `true` if complete.
      #
      def complete?
        status == :complete
      end

      # [public] Return `true` if failed or complete.
      #
      def resolved?
        failed? || complete?
      end

      private def resolve_value(value)
        if value.is_a?(self.class)
          resolve_value(value.result)
        else
          value
        end
      end

      private def wait_all(task)
        task.children&.each do |child|
          wait_all(child)
        end

        task.wait
      end
    end
  end
end
