# frozen_string_literal: true

require_relative "../core/async/bootstrap"
require_relative "../core/async/future"
require_relative "../core/async/timeout"
require_relative "../core/async/wrapper"

module Is
  # [public] Makes objects async-aware.
  #
  module Async
    # [public] Call behavior asychronously, returning a future.
    #
    def async
      if block_given?
        task = ::Async::Reactor.run { |current|
          begin
            yield
          ensure
            current.yield
          end
        }

        Core::Async::Future.new(task)
      else
        Core::Async::Wrapper.new(self)
      end
    end

    # [public] Call behavior synchronously but within an async context, waiting on the result.
    #
    private def await
      if (task = ::Async::Task.current?)
        reference = task.async { |current|
          begin
            yield
          ensure
            current.yield
          end
        }

        wait_all(reference)
      else
        ::Async::Reactor.run { |task|
          async {
            yield
          }.result
        }.wait
      end
    rescue UncaughtThrowError => error
      throw error.tag, error.value
    end

    # [public] Call behavior within an async context without additional nesting.
    #
    private def aware
      if ::Async::Task.current?
        yield
      else
        await do
          yield
        end
      end
    end

    # [public] Sleeps for `seconds` in a proper async context.
    #
    private def sleep(seconds)
      internal_await do |task|
        task.sleep(seconds)
      end
    end

    # [public] Call asynchonous behavior in a proper async context, wrapped in a timeout.
    #
    # Raises `Core::Async::Timeout` if execution exceeds `seconds`.
    #
    private def timeout(seconds)
      internal_await do |task|
        timed_task = internal_async {
          yield
        }

        if seconds && seconds > 0
          task.with_timeout(seconds, Core::Async::Timeout) do
            timed_task.wait
          end
        else
          timed_task.wait
        end
      ensure
        timed_task&.stop
      end
    rescue UncaughtThrowError => error
      throw error.tag, error.value
    end

    # [public] Yields control to allow other fibers to execute.
    #
    private def defer
      if (task = ::Async::Task.current?)
        task.yield
      end
    end

    # [public] Cancels the current async behavior if in progress.
    #
    private def cancel
      if (task = ::Async::Task.current?)
        task.stop
      end
    end

    # [public] Resolves a potential future to a final result.
    #
    private def resolve(value)
      case value
      when Core::Async::Future
        value.result
      else
        value
      end
    end

    private def internal_async
      ::Async::Reactor.run do |task|
        yield task
      end
    end

    private def internal_await
      if (task = ::Async::Task.current?)
        reference = task.async {
          yield task
        }

        wait_all(reference)
      else
        ::Async::Reactor.run { |task|
          yield task
        }.wait
      end
    end

    private def wait_all(task)
      task.children&.each do |child|
        wait_all(child)
      end

      task.wait
    end
  end
end
